import logging

import numpy as np
import pytest
from sklearn.linear_model import LinearRegression

from pydvl.utils import GroupedDataset, MemcachedConfig, Status, Utility
from pydvl.utils.numeric import num_samples_permutation_hoeffding
from pydvl.utils.score import Scorer, squashed_r2
from pydvl.value import compute_shapley_values
from pydvl.value.shapley import ShapleyMode
from pydvl.value.shapley.naive import combinatorial_exact_shapley
from pydvl.value.stopping import MaxChecks, MaxUpdates

from .. import check_rank_correlation, check_total_value, check_values

log = logging.getLogger(__name__)


# noinspection PyTestParametrized
@pytest.mark.parametrize(
    "num_samples, fun, rtol, atol, kwargs",
    [
        (12, ShapleyMode.PermutationMontecarlo, 0.1, 1e-5, {"done": MaxUpdates(10)}),
        # FIXME! it should be enough with 2**(len(data)-1) samples
        (
            8,
            ShapleyMode.CombinatorialMontecarlo,
            0.2,
            1e-4,
            {"done": MaxUpdates(2**10)},
        ),
        (12, ShapleyMode.Owen, 0.1, 1e-4, dict(n_samples=4, max_q=200)),
        (12, ShapleyMode.OwenAntithetic, 0.1, 1e-4, dict(n_samples=4, max_q=200)),
        (
            4,
            ShapleyMode.GroupTesting,
            0.1,
            # Because of the inaccuracy of GTS, a high atol is required for the
            # value 0, for which the rtol has no effect.
            1e-2,
            dict(n_samples=int(4e5), epsilon=0.2, delta=0.01),
        ),
    ],
)
def test_analytic_montecarlo_shapley(
    num_samples,
    analytic_shapley,
    parallel_config,
    n_jobs,
    fun: ShapleyMode,
    rtol: float,
    atol: float,
    kwargs: dict,
):
    u, exact_values = analytic_shapley

    values = compute_shapley_values(
        u, mode=fun, n_jobs=n_jobs, config=parallel_config, progress=False, **kwargs
    )

    check_values(values, exact_values, rtol=rtol, atol=atol)


@pytest.mark.parametrize("num_samples, delta, eps", [(8, 0.1, 0.1)])
@pytest.mark.parametrize(
    "fun", [ShapleyMode.PermutationMontecarlo, ShapleyMode.CombinatorialMontecarlo]
)
def test_hoeffding_bound_montecarlo(
    num_samples,
    analytic_shapley,
    tolerate,
    n_jobs,
    fun: ShapleyMode,
    delta: float,
    eps: float,
):
    u, exact_values = analytic_shapley

    n_samples = num_samples_permutation_hoeffding(delta=delta, eps=eps, u_range=1)

    for _ in range(10):
        with tolerate(max_failures=int(10 * delta)):
            values = compute_shapley_values(
                u=u, mode=fun, done=MaxChecks(n_samples), n_jobs=n_jobs
            )
            # Trivial bound on total error using triangle inequality
            check_total_value(u, values, atol=len(u.data) * eps)
            check_rank_correlation(values, exact_values, threshold=0.8)


@pytest.mark.parametrize(
    "a, b, num_points", [(2, 0, 21)]  # training set will have 0.3 * 21 = 6 samples
)
@pytest.mark.parametrize("scorer, rtol", [(squashed_r2, 0.25)])
@pytest.mark.parametrize(
    "fun, kwargs",
    [
        # FIXME: Hoeffding says 400 should be enough
        (ShapleyMode.PermutationMontecarlo, dict(done=MaxUpdates(600))),
        (ShapleyMode.CombinatorialMontecarlo, dict(done=MaxUpdates(2**11))),
        (ShapleyMode.Owen, dict(n_samples=2, max_q=300)),
        (ShapleyMode.OwenAntithetic, dict(n_samples=2, max_q=300)),
        (
            ShapleyMode.GroupTesting,
            dict(n_samples=int(1e5), epsilon=0.2, delta=0.01),
        ),
    ],
)
def test_linear_montecarlo_shapley(
    linear_shapley,
    n_jobs,
    memcache_client_config,
    scorer: Scorer,
    rtol: float,
    fun: ShapleyMode,
    kwargs: dict,
):
    """Tests values for all methods using a linear dataset.

    For permutation and truncated montecarlo, the rtol for each scorer is chosen
    so that the number of samples selected is just above the (ε,δ) bound for ε =
    rtol, δ=0.001 and the range corresponding to each score. This means that
    roughly once every 1000/num_methods runs the test will fail.

    FIXME:
     - For permutation, we must increase the number of samples above that what
       is done for truncated, this is probably due to the averaging done by the
       latter to reduce variance
     - We don't have a bound for Owen.
    NOTE:
     - The variance in the combinatorial method is huge, so we need lots of
       samples

    """
    u, exact_values = linear_shapley
    check_total_value(u, exact_values, rtol=rtol)

    values = compute_shapley_values(
        u, mode=fun, progress=False, n_jobs=n_jobs, **kwargs
    )

    check_values(values, exact_values, rtol=rtol)
    check_total_value(u, values, rtol=rtol)  # FIXME, could be more than rtol


@pytest.mark.parametrize(
    "a, b, num_points", [(2, 0, 21)]  # training set will have 0.3 * 21 ~= 6 samples
)
@pytest.mark.parametrize("scorer, total_atol", [(squashed_r2, 0.2)])
@pytest.mark.parametrize(
    "fun, kwargs",
    [
        (ShapleyMode.PermutationMontecarlo, {"done": MaxUpdates(500)}),
        (ShapleyMode.Owen, dict(n_samples=4, max_q=400)),
        (ShapleyMode.OwenAntithetic, dict(n_samples=4, max_q=400)),
        (
            ShapleyMode.GroupTesting,
            dict(n_samples=int(1e5), epsilon=0.2, delta=0.01),
        ),
    ],
)
def test_linear_montecarlo_with_outlier(
    linear_dataset,
    n_jobs,
    memcache_client_config,
    scorer: Scorer,
    total_atol: float,
    fun,
    kwargs: dict,
):
    """Tests whether valuation methods are able to detect an obvious outlier.

    A point is selected at random from a linear dataset and the dependent
    variable is set to 10 standard deviations.

    Note that this implies that the whole dataset will have very low utility:
    e.g. for R^2 it will be very negative. The larger the range of the utility,
    the more samples are required for the Monte Carlo approximations to converge,
    as indicated by the Hoeffding bound.
    """
    outlier_idx = np.random.randint(len(linear_dataset.y_train))
    linear_dataset.y_train[outlier_idx] = np.std(linear_dataset.y_train) * 10
    linear_utility = Utility(
        LinearRegression(),
        data=linear_dataset,
        scorer=scorer,
        cache_options=MemcachedConfig(client_config=memcache_client_config),
    )
    values = compute_shapley_values(
        linear_utility, mode=fun, progress=False, n_jobs=n_jobs, **kwargs
    )
    values.sort()

    assert values.status == Status.Converged
    check_total_value(linear_utility, values, atol=total_atol)
    assert values[0].index == outlier_idx


@pytest.mark.parametrize(
    "a, b, num_points, num_groups", [(2, 0, 21, 2)]  # 24*0.3=6 samples in 2 groups
)
@pytest.mark.parametrize("scorer, rtol", [(squashed_r2, 0.1)])
@pytest.mark.parametrize(
    "fun, kwargs",
    [
        (ShapleyMode.PermutationMontecarlo, dict(done=MaxUpdates(700))),
    ],
)
def test_grouped_linear_montecarlo_shapley(
    linear_dataset,
    n_jobs,
    memcache_client_config: "MemcachedClientConfig",
    num_groups: int,
    fun: ShapleyMode,
    scorer: Scorer,
    rtol: float,
    kwargs: dict,
):
    """
    For permutation and truncated montecarlo, the rtol for each scorer is chosen
    so that the number of samples selected is just above the (ε,δ) bound for ε =
    rtol, δ=0.001 and the range corresponding to each score. This means that
    roughly once every 1000/num_methods runs the test will fail.
    """
    data_groups = np.random.randint(0, num_groups, len(linear_dataset))
    grouped_linear_dataset = GroupedDataset.from_dataset(linear_dataset, data_groups)
    grouped_linear_utility = Utility(
        LinearRegression(),
        data=grouped_linear_dataset,
        scorer=scorer,
        cache_options=MemcachedConfig(client_config=memcache_client_config),
    )
    exact_values = combinatorial_exact_shapley(grouped_linear_utility, progress=False)

    values = compute_shapley_values(
        grouped_linear_utility, mode=fun, progress=False, n_jobs=n_jobs, **kwargs
    )

    check_values(values, exact_values, rtol=rtol)



import operator
import os
import time
from functools import partial, reduce

import numpy as np
import pytest

from pydvl.utils.parallel import MapReduceJob, init_parallel_backend
from pydvl.utils.parallel.backend import available_cpus, effective_n_jobs
from pydvl.utils.parallel.futures import init_executor
from pydvl.utils.parallel.map_reduce import _get_value


def test_effective_n_jobs(parallel_config, num_workers):
    parallel_backend = init_parallel_backend(parallel_config)
    if parallel_config.backend == "sequential":
        assert parallel_backend.effective_n_jobs(1) == 1
        assert parallel_backend.effective_n_jobs(4) == 1
        assert parallel_backend.effective_n_jobs(-1) == 1
    else:
        assert parallel_backend.effective_n_jobs(1) == 1
        assert parallel_backend.effective_n_jobs(4) == 4
        if parallel_config.address is None:
            assert parallel_backend.effective_n_jobs(-1) == num_workers
        else:
            assert parallel_backend.effective_n_jobs(-1) == num_workers

    for n_jobs in [-1, 1, 2]:
        assert parallel_backend.effective_n_jobs(n_jobs) == effective_n_jobs(
            n_jobs, parallel_config
        )
        assert effective_n_jobs(n_jobs, parallel_config) > 0

    with pytest.raises(ValueError):
        parallel_backend.effective_n_jobs(0)


@pytest.fixture()
def map_reduce_job_and_parameters(parallel_config, n_jobs, request):
    try:
        kind, map_func, reduce_func = request.param
        assert kind == "custom"
    except ValueError:
        kind = request.param
    if kind == "numpy":
        map_reduce_job = partial(
            MapReduceJob,
            map_func=np.sum,
            reduce_func=np.sum,
            config=parallel_config,
            n_jobs=n_jobs,
        )
    elif kind == "list":
        map_reduce_job = partial(
            MapReduceJob,
            map_func=lambda x: x,
            reduce_func=lambda r: reduce(operator.add, r, []),
            config=parallel_config,
            n_jobs=n_jobs,
        )
    elif kind == "range":
        map_reduce_job = partial(
            MapReduceJob,
            map_func=lambda x: list(x),
            reduce_func=lambda r: reduce(operator.add, list(r), []),
            config=parallel_config,
            n_jobs=n_jobs,
        )
    elif kind == "custom":
        map_reduce_job = partial(
            MapReduceJob,
            map_func=map_func,
            reduce_func=reduce_func,
            config=parallel_config,
            n_jobs=n_jobs,
        )
    else:
        map_reduce_job = partial(
            MapReduceJob,
            map_func=lambda x: x * x,
            reduce_func=lambda r: r,
            config=parallel_config,
            n_jobs=n_jobs,
        )
    return map_reduce_job, n_jobs


@pytest.mark.parametrize(
    "map_reduce_job_and_parameters, indices, expected",
    [
        ("list", [], []),
        ("list", [1, 2], [1, 2]),
        ("list", [1, 2, 3, 4], [1, 2, 3, 4]),
        ("range", range(10), list(range(10))),
        ("numpy", np.arange(10), 45),
    ],
    indirect=["map_reduce_job_and_parameters"],
)
@pytest.mark.parametrize("n_jobs", [1, 2, 4])
def test_map_reduce_job(map_reduce_job_and_parameters, indices, expected):
    map_reduce_job, n_jobs = map_reduce_job_and_parameters
    result = map_reduce_job(indices)()
    if not isinstance(result, np.ndarray):
        assert result == expected
    else:
        assert (result == expected).all()


@pytest.mark.parametrize(
    "data, n_chunks, expected_chunks",
    [
        ([], 3, []),
        ([1, 2, 3], 2, [[1, 2], [3]]),
        ([1, 2, 3, 4], 2, [[1, 2], [3, 4]]),
        ([1, 2, 3, 4], 3, [[1, 2], [3], [4]]),
        ([1, 2, 3, 4], 5, [[1], [2], [3], [4]]),
        (list(range(5)), 42, [[i] for i in range(5)]),
        (np.arange(5), 42, [[i] for i in range(5)]),
        (range(10), 4, [range(0, 3), range(3, 6), range(6, 8), range(8, 10)]),
        (np.arange(10), 4, np.array_split(np.arange(10), 4)),
    ],
)
def test_chunkification(data, n_chunks, expected_chunks):
    map_reduce_job = MapReduceJob([], map_func=lambda x: x)
    chunks = list(map_reduce_job._chunkify(data, n_chunks))
    chunks = map_reduce_job.parallel_backend.get(chunks)
    for x, y in zip(chunks, expected_chunks):
        if not isinstance(x, np.ndarray):
            assert x == y
        else:
            assert (x == y).all()


@pytest.mark.parametrize(
    "max_parallel_tasks, n_finished, n_dispatched, expected_n_finished",
    [
        (1, 3, 6, 5),
        (3, 3, 3, 3),
        (10, 1, 15, 5),
        (20, 1, 3, 1),
    ],
)
def test_backpressure(
    max_parallel_tasks, n_finished, n_dispatched, expected_n_finished
):
    def map_func(x):
        import time

        time.sleep(1)
        return x

    inputs_ = list(range(n_dispatched))

    map_reduce_job = MapReduceJob(
        inputs_,
        map_func=map_func,
        max_parallel_tasks=max_parallel_tasks,
        timeout=10,
    )

    map_func = map_reduce_job._wrap_function(map_func)
    jobs = [map_func(x) for x in inputs_]
    n_finished = map_reduce_job._backpressure(
        jobs, n_finished=n_finished, n_dispatched=n_dispatched
    )
    assert n_finished == expected_n_finished


def test_map_reduce_job_partial_map_and_reduce_func(parallel_config):
    def map_func(x, y):
        return x + y

    def reduce_func(x, y):
        return np.sum(np.concatenate(x)) + y

    map_func = partial(map_func, y=10)
    reduce_func = partial(reduce_func, y=5)

    map_reduce_job = MapReduceJob(
        np.arange(10),
        map_func=map_func,
        reduce_func=reduce_func,
        config=parallel_config,
    )
    result = map_reduce_job()
    assert result == 150


@pytest.mark.parametrize(
    "x, expected_x",
    [
        (None, None),
        ([0, 1], [0, 1]),
        (np.arange(3), np.arange(3)),
    ],
)
def test_map_reduce_get_value(x, expected_x, parallel_config):
    assert np.all(_get_value(x) == expected_x)
    parallel_backend = init_parallel_backend(parallel_config)
    x_id = parallel_backend.put(x)
    assert np.all(_get_value(x_id) == expected_x)


def test_wrap_function(parallel_config, num_workers):
    def fun(x, **kwargs):
        return dict(x=x * x, **kwargs)

    parallel_backend = init_parallel_backend(parallel_config)
    # Try two kwargs for @ray.remote. Should be ignored in the sequential backend
    wrapped_func = parallel_backend.wrap(fun, num_cpus=1, max_calls=1)
    x = parallel_backend.put(2)
    ret = parallel_backend.get(wrapped_func(x))

    assert ret["x"] == 4
    assert len(ret) == 1  # Ensure that kwargs are not passed to the function

    if parallel_config.backend != "sequential":
        # Test that the function is executed in different processes
        def get_pid():
            time.sleep(2)  # FIXME: waiting less means fewer processes are used?!
            return os.getpid()

        wrapped_func = parallel_backend.wrap(get_pid, num_cpus=1)
        pids = parallel_backend.get([wrapped_func() for _ in range(num_workers)])
        assert len(set(pids)) == num_workers


def test_futures_executor_submit(parallel_config):
    with init_executor(config=parallel_config) as executor:
        future = executor.submit(lambda x: x + 1, 1)
        result = future.result()
    assert result == 2


def test_futures_executor_map(parallel_config):
    with init_executor(config=parallel_config) as executor:
        results = list(executor.map(lambda x: x + 1, range(3)))
    assert results == [1, 2, 3]


def test_futures_executor_map_with_max_workers(parallel_config, num_workers):
    if parallel_config.backend != "ray":
        pytest.skip("Currently this test only works with Ray")

    def func(_):
        time.sleep(1)
        return time.monotonic()

    start_time = time.monotonic()
    with init_executor(config=parallel_config) as executor:
        assert executor._max_workers == num_workers
        list(executor.map(func, range(3)))
    end_time = time.monotonic()
    total_time = end_time - start_time
    # We expect the time difference to be > 3 / num_workers, but has to be at least 1
    assert total_time > max(1.0, 3 / num_workers)



"""
Contains

- batched conjugate gradient.
- error bound for conjugate gradient.
"""
import logging
import warnings
from typing import TYPE_CHECKING, Callable, Optional, Tuple, Union

import numpy as np
from scipy.sparse.linalg import cg

from ..utils import maybe_progress
from .types import MatrixVectorProduct

if TYPE_CHECKING:
    from numpy.typing import NDArray

__all__ = ["conjugate_gradient", "batched_preconditioned_conjugate_gradient"]

logger = logging.getLogger(__name__)


def conjugate_gradient(
    A: "NDArray[np.float_]", batch_y: "NDArray[np.float_]", progress: bool = False
) -> "NDArray[np.float_]":
    """
    Given a matrix and a batch of vectors, it uses conjugate gradient to calculate the solution
    to Ax = y for each y in batch_y.

    :param A: a real, symmetric and positive-definite matrix of shape [NxN]
    :param batch_y: a matrix of shape [NxP], with P the size of the batch.
    :param progress: True, iff progress shall be printed.

    :return: A NDArray of shape [NxP] representing x, the solution of Ax=b.
    """
    batch_cg = []
    for y in maybe_progress(batch_y, progress, desc="Conjugate gradient"):
        y_cg, _ = cg(A, y)
        batch_cg.append(y_cg)
    return np.asarray(batch_cg)


def batched_preconditioned_conjugate_gradient(
    A: Union["NDArray", Callable[["NDArray"], "NDArray"]],
    b: "NDArray",
    x0: Optional["NDArray"] = None,
    rtol: float = 1e-3,
    max_iterations: int = 100,
    max_step_size: Optional[float] = None,
) -> Tuple["NDArray", int]:
    """
    Implementation of a batched conjugate gradient algorithm. It uses vector matrix products for efficient calculation.
    On top of that, it constrains the maximum step size.

    See [1]_ for more details on the algorithm.

    See also [2]_ and [3]_.

    .. warning::

        This function is experimental and unstable. Prefer using inversion_method='cg'

    :param A: A linear function f : R[k] -> R[k] representing a matrix vector product from dimension K to K or a matrix. \
        It has to be positive-definite v.T @ f(v) >= 0.
    :param b: A NDArray of shape [K] representing the targeted result of the matrix multiplication Ax.
    :param max_iterations: Maximum number of iterations to use in conjugate gradient. Default is 10 times K.
    :param rtol: Relative tolerance of the residual with respect to the 2-norm of b.
    :param max_step_size: Maximum step size along a gradient direction. Might be necessary for numerical stability. \
        See also max_iterations. Default is 10.0.
    :param verify_assumptions: True, iff the matrix should be checked for positive-definiteness by a stochastic rule.

    :return: A NDArray of shape [K] representing the solution of Ax=b.

    .. note::
        .. [1] `Conjugate Gradient Method - Wikipedia <https://en.wikipedia.org/wiki/Conjugate_gradient_method>`_.
        .. [2] `SciPy's implementation of Conjugate Gradient <https://github.com/scipy/scipy/blob/v1.8.1/scipy/sparse/linalg/_isolve/iterative.py#L282-L351>`_.
        .. [3] `Prof. Mert Pilanci., "Conjugate Gradient Method", Stanford University, 2022 <https://web.stanford.edu/class/ee364b/lectures/conj_grad_slides.pdf>`_.
    """
    warnings.warn(
        "This function is experimental and unstable. Prefer using inversion_method='cg'",
        UserWarning,
    )
    # wrap A into a function.
    if not callable(A):
        new_A = np.copy(A)
        A = lambda v: v @ new_A.T  # type: ignore
    M = hvp_to_inv_diag_conditioner(A, d=b.shape[1])

    k = A(b).shape[0]
    if A(b).size == 0:
        return b, 0

    if b.ndim == 1:
        b = b.reshape([1, -1])

    if max_iterations is None:
        max_iterations = 10 * k

    # start with residual
    if x0 is not None:
        x = np.copy(x0)
    elif M is not None:
        x = M(b)
    else:
        x = np.copy(b)

    r = b - A(x)
    u = np.copy(r)

    if M is not None:
        u = M(u)

    p = np.copy(b)

    if x.ndim == 1:
        x = x.reshape([1, -1])

    iteration = 0
    batch_dim = b.shape[0]
    converged = np.zeros(batch_dim, dtype=bool)
    atol = np.linalg.norm(b, axis=1) * rtol

    while iteration < max_iterations:
        # remaining fields
        iteration += 1
        not_yet_converged_indices = np.argwhere(np.logical_not(converged))[:, 0]
        mvp = A(p)[not_yet_converged_indices]
        p_dot_mvp = np.einsum("ia,ia->i", p[not_yet_converged_indices], mvp)
        r_dot_u = np.einsum(
            "ia,ia->i", r[not_yet_converged_indices], u[not_yet_converged_indices]
        )
        alpha = r_dot_u / p_dot_mvp
        if max_step_size is not None:
            alpha = np.minimum(max_step_size, alpha)

        # update x and r
        reshaped_alpha = alpha.reshape([-1, 1])
        x[not_yet_converged_indices] += reshaped_alpha * p[not_yet_converged_indices]
        r[not_yet_converged_indices] -= reshaped_alpha * mvp

        # calculate next conjugate gradient
        new_u = r
        if M is not None:
            new_u = M(new_u)

        new_u = new_u[not_yet_converged_indices]
        new_r_dot_u = np.einsum("ia,ia->i", r[not_yet_converged_indices], new_u)

        if rtol is not None:
            residual = np.linalg.norm(
                A(x)[not_yet_converged_indices] - b[not_yet_converged_indices],
                axis=1,
            )
            converged[not_yet_converged_indices] = (
                residual <= atol[not_yet_converged_indices]
            )

        if np.all(converged):
            break

        beta = new_r_dot_u / r_dot_u
        p[not_yet_converged_indices] = (
            beta.reshape([-1, 1]) * p[not_yet_converged_indices] + new_u
        )
        u[not_yet_converged_indices] = new_u

    if not np.all(converged):
        percentage_converged = int(converged.sum() / len(converged)) * 100
        msg = (
            f"Converged vectors are only {percentage_converged}%. "
            "Please increase max_iterations, decrease max_step_size "
            "and make sure that A is positive definite"
            " (e.g. through regularization)."
        )
        warnings.warn(msg, RuntimeWarning)
    return x, iteration


def conjugate_gradient_condition_number_based_error_bound(
    A: "NDArray", n: int, x0: "NDArray", xt: "NDArray"
) -> float:
    """
    Error bound for conjugate gradient based on the condition number of the weight matrix A. Used for testing purposes.
    See also https://math.stackexchange.com/questions/382958/error-for-conjugate-gradient-method. Explicit of the weight
    matrix is required.
    :param A: Weight matrix of the matrix to be inverted.
    :param n: Maximum number for executed iterations X in conjugate gradient.
    :param x0: Initialization solution x0 of conjugate gradient.
    :param xt: Final solution xt of conjugate gradient after X iterations.
    :returns: Upper bound for ||x0 - xt||_A.
    """
    eigvals = np.linalg.eigvals(A)
    eigvals = np.sort(eigvals)
    eig_val_max = np.max(eigvals)
    eig_val_min = np.min(eigvals)
    kappa = np.abs(eig_val_max / eig_val_min)
    norm_A = lambda v: np.sqrt(np.einsum("ia,ab,ib->i", v, A, v))
    error_init: float = norm_A(xt - x0)

    sqrt_kappa = np.sqrt(kappa)
    div = (sqrt_kappa + 1) / (sqrt_kappa - 1)
    div_n = div**n
    return (2 * error_init) / (div_n + 1 / div_n)  # type: ignore


def hvp_to_inv_diag_conditioner(
    hvp: MatrixVectorProduct, d: int
) -> MatrixVectorProduct:
    """
    This method uses the hvp function to construct a simple pre-conditioner 1/diag(H). It does so while requiring
    only O(d) space in RAM for construction and later execution.
    :param hvp: The callable calculating the Hessian vector product Hv.
    :param d: The number of dimensions of the hvp callable.
    :returns: A MatrixVectorProduct for the conditioner.
    """
    diags = np.empty(d)

    for i in range(d):
        inp = np.zeros(d)
        inp[i] = 1
        diags[i] = hvp(np.reshape(inp, [1, -1]))[0, i]

    def _inv_diag_conditioner(v: "NDArray"):
        return v / diags

    return _inv_diag_conditioner



import numpy as np
from numpy.typing import NDArray

from pydvl.utils.score import Scorer, compose_score, squashed_r2, squashed_variance

sigmoid = lambda x: 1 / (1 + np.exp(-x))


class FittedLinearModel:
    def __init__(self, coef: NDArray):
        self.coef = coef

    def predict(self, X):
        return X @ self.coef

    def fit(self, X, y):
        pass

    def score(self, X, y):
        return np.linalg.norm(X @ self.coef - y)


def test_scorer():
    """Tests the Scorer class."""
    scorer = Scorer("r2")
    assert str(scorer) == "r2"
    assert repr(scorer) == "R2 (scorer=make_scorer(r2_score))"

    coef = np.array([1, 2])
    X = np.array([[1, 2], [3, 4]])
    model = FittedLinearModel(coef)
    assert 1.0 == scorer(model, X, X @ coef)


def test_compose_score():
    """Tests the compose_score function."""
    composed = compose_score(Scorer("r2"), sigmoid, (0, 1), "squashed r2")
    assert str(composed) == "squashed r2"
    assert repr(composed) == "SquashedR2 (scorer=r2)"

    coef = np.array([1, 2])
    X = np.array([[1, 2], [3, 4]])
    model = FittedLinearModel(coef)
    assert sigmoid(1.0) == composed(model, X, X @ coef)


def test_squashed_r2():
    """Tests the squashed_r2 scorer."""
    assert str(squashed_r2) == "squashed r2"
    assert repr(squashed_r2) == "SquashedR2 (scorer=r2)"
    assert np.allclose(squashed_r2.range, (0, 1))

    coef = np.array([1, 2])
    X = np.array([[1, 2], [3, 4]])
    model = FittedLinearModel(coef)
    assert sigmoid(1.0) == squashed_r2(model, X, X @ coef)


def test_squashed_variance():
    """Tests the squashed_variance scorer."""
    assert str(squashed_variance) == "squashed explained variance"
    assert (
        repr(squashed_variance)
        == "SquashedExplainedVariance (scorer=explained_variance)"
    )
    assert np.allclose(squashed_variance.range, (0, 1))

    coef = np.array([1, 2])
    X = np.array([[1, 2], [3, 4]])
    model = FittedLinearModel(coef)
    assert sigmoid(1.0) == squashed_variance(model, X, X @ coef)



from typing import Tuple

import numpy as np
import pytest

from pydvl.utils import Utility
from pydvl.utils.status import Status
from pydvl.utils.utility import GlovesGameUtility, MinerGameUtility
from pydvl.value.result import ValuationResult


@pytest.fixture(scope="module")
def test_utility(request) -> Tuple[Utility, ValuationResult]:
    name, kwargs = request.param
    if name == "miner":
        u = MinerGameUtility(**kwargs)
    elif name == "gloves":
        u = GlovesGameUtility(**kwargs)
    else:
        raise ValueError(f"Unknown '{name}'")
    exact_values, subsidy = u.exact_least_core_values()
    result = ValuationResult(
        algorithm="exact",
        values=exact_values,
        subsidy=subsidy,
        variances=np.zeros_like(exact_values),
        data_names=np.arange(len(exact_values)),
        status=Status.Converged,
    )
    return u, result



r"""
Algorithms for the exact and approximate computation of value and semi-value.

See :ref:`data valuation` for an introduction to the concepts and methods
implemented here.
"""

from .result import *  # isort: skip
from ..utils import Dataset, Scorer, Utility
from .least_core import *
from .loo import *
from .sampler import *
from .semivalues import *
from .shapley import *
from .stopping import *



r"""
This module provides the core functionality for the computation of generic
semi-values. A **semi-value** is any valuation function with the form:

$$v_\text{semi}(i) = \sum_{i=1}^n w(k)
\sum_{S \subset D_{-i}^{(k)}} [U(S_{+i})-U(S)],$$

where the coefficients $w(k)$ satisfy the property:

$$\sum_{k=1}^n w(k) = 1.$$

As such, the computation of a semi-value requires two components:

1. A **subset sampler** that generates subsets of the set $D$ of interest.
2. A **coefficient** $w(k)$ that assigns a weight to each subset size $k$.

Samplers can be found in :mod:`pydvl.value.sampler`, and can be classified into
two categories: powerset samplers and (one) permutation sampler. Powerset
samplers generate subsets of $D_{-i}$, while the permutation sampler generates
permutations of $D$. The former conform to the above definition of semi-values,
while the latter reformulates it as:

$$
v_u(x_i) = \frac{1}{n!} \sum_{\sigma \in \Pi(n)}
\tilde{w}( | \sigma_{:i} | )[u(\sigma_{:i} \cup \{i\}) − u(\sigma_{:i})],
$$

where $\sigma_{:i}$ denotes the set of indices in permutation sigma before the
position where $i$ appears (see :ref:`data valuation` for details), and
$\tilde{w}(k) = n \choose{n-1}{k} w(k)$ is the weight correction due to the
reformulation.


There are several pre-defined coefficients, including the Shapley value
of :footcite:t:`ghorbani_data_2019`, the Banzhaf index of
:footcite:t:`wang_data_2022`, and the Beta coefficient of
:footcite:t:`kwon_beta_2022`.

.. note::
   For implementation consistency, we slightly depart from the common definition
   of semi-values, which includes a factor $1/n$ in the sum over subsets.
   Instead, we subsume this factor into the coefficient $w(k)$.

"""

import math
import operator
from enum import Enum
from functools import reduce
from itertools import takewhile
from typing import Protocol, Type, cast

import scipy as sp
from tqdm import tqdm

from pydvl.utils import MapReduceJob, ParallelConfig, Utility
from pydvl.value import ValuationResult
from pydvl.value.sampler import PermutationSampler, PowersetSampler
from pydvl.value.stopping import MaxUpdates, StoppingCriterion

__all__ = [
    "banzhaf_coefficient",
    "beta_coefficient",
    "shapley_coefficient",
    "semivalues",
    "SemiValueMode",
]


class SVCoefficient(Protocol):
    """A coefficient for the computation of semi-values."""

    __name__: str

    def __call__(self, n: int, k: int) -> float:
        """Computes the coefficient for a given subset size.

        :param n: Total number of elements in the set.
        :param k: Size of the subset for which the coefficient is being computed
        """
        ...


def _semivalues(
    sampler: PowersetSampler,
    u: Utility,
    coefficient: SVCoefficient,
    done: StoppingCriterion,
    *,
    progress: bool = False,
    job_id: int = 1,
) -> ValuationResult:
    r"""Serial computation of semi-values. This is a helper function for
    :func:`semivalues`.

    :param sampler: The subset sampler to use for utility computations.
    :param u: Utility object with model, data, and scoring function.
    :param coefficient: The semivalue coefficient
    :param done: Stopping criterion.
    :param progress: Whether to display progress bars for each job.
    :param job_id: id to use for reporting progress.
    :return: Object with the results.
    """
    n = len(u.data.indices)
    result = ValuationResult.zeros(
        algorithm=f"semivalue-{str(sampler)}-{coefficient.__name__}",
        indices=sampler.indices,
        data_names=[u.data.data_names[i] for i in sampler.indices],
    )

    samples = takewhile(lambda _: not done(result), sampler)
    pbar = tqdm(disable=not progress, position=job_id, total=100, unit="%")
    for idx, s in samples:
        pbar.n = 100 * done.completion()
        pbar.refresh()
        marginal = (
            (u({idx}.union(s)) - u(s)) * coefficient(n, len(s)) * sampler.weight(s)
        )
        result.update(idx, marginal)

    return result


def semivalues(
    sampler: PowersetSampler,
    u: Utility,
    coefficient: SVCoefficient,
    done: StoppingCriterion,
    *,
    n_jobs: int = 1,
    config: ParallelConfig = ParallelConfig(),
    progress: bool = False,
) -> ValuationResult:
    """
    Computes semi-values for a given utility function and subset sampler.

    :param sampler: The subset sampler to use for utility computations.
    :param u: Utility object with model, data, and scoring function.
    :param coefficient: The semi-value coefficient
    :param done: Stopping criterion.
    :param n_jobs: Number of parallel jobs to use.
    :param config: Object configuring parallel computation, with cluster
        address, number of cpus, etc.
    :param progress: Whether to display progress bars for each job.
    :return: Object with the results.

    """
    map_reduce_job: MapReduceJob[PowersetSampler, ValuationResult] = MapReduceJob(
        sampler,
        map_func=_semivalues,
        reduce_func=lambda results: reduce(operator.add, results),
        map_kwargs=dict(u=u, coefficient=coefficient, done=done, progress=progress),
        config=config,
        n_jobs=n_jobs,
    )
    return map_reduce_job()


def shapley_coefficient(n: int, k: int) -> float:
    return float(1 / math.comb(n - 1, k) / n)


def banzhaf_coefficient(n: int, k: int) -> float:
    return float(1 / 2 ** (n - 1))


def beta_coefficient(alpha: float, beta: float) -> SVCoefficient:
    # Leave this here in case we want to avoid depending on scipy
    # n = n or 1024
    # @lru_cache(maxsize=int(n * (n - 1) / 2))
    # def B(a: int, b: int) -> float:
    #     """Γ(a) * Γ(b) / Γ(a+b)"""
    #     ii = np.arange(1, b)
    #     return (ii / (a + ii)).prod() / a

    B = sp.special.beta
    const = B(alpha, beta)

    def beta_coefficient_w(n: int, k: int) -> float:
        """Beta coefficient"""
        j = k + 1
        w = B(j + beta - 1, n - j + alpha) / const
        # return math.comb(n - 1, j - 1) * w * n
        return float(w)

    return cast(SVCoefficient, beta_coefficient_w)


class SemiValueMode(str, Enum):
    Shapley = "shapley"
    BetaShapley = "beta_shapley"
    Banzhaf = "banzhaf"


def compute_semivalues(
    u: Utility,
    *,
    done: StoppingCriterion = MaxUpdates(100),
    mode: SemiValueMode = SemiValueMode.Shapley,
    sampler_t: Type[PowersetSampler] = PermutationSampler,
    n_jobs: int = 1,
    **kwargs,
) -> ValuationResult:
    """Entry point for most common semi-value computations. All are implemented
    with permutation sampling.

    For any other sampling method, use :func:`parallel_semivalues` directly.

    See :ref:`data valuation` for an overview of valuation.

    The modes supported are:

    - :attr:`SemiValueMode.Shapley`: Shapley values.
    - :attr:`SemiValueMode.BetaShapley`: Implements the Beta Shapley semi-value
        as introduced in :footcite:t:`kwon_beta_2022`. Pass additional keyword
        arguments ``alpha`` and ``beta`` to set the parameters of the Beta
        distribution (both default to 1).
    - :attr:`SemiValueMode.Banzhaf`: Implements the Banzhaf semi-value as
        introduced in :footcite:t:`wang_data_2022`.

    :param u: Utility object with model, data, and scoring function.
    :param done: Stopping criterion.
    :param mode: The semi-value mode to use. See :class:`SemiValueMode` for a
        list.
    :param sampler_t: The sampler type to use. See :mod:`pydvl.value.sampler`
        for a list.
    :param n_jobs: Number of parallel jobs to use.
    :param kwargs: Additional keyword arguments passed to
        :func:`~pydvl.value.semivalues.semivalues`.
    """
    sampler_instance = sampler_t(u.data.indices)
    if mode == SemiValueMode.Shapley:
        coefficient = shapley_coefficient
    elif mode == SemiValueMode.BetaShapley:
        alpha = kwargs.pop("alpha", 1)
        beta = kwargs.pop("beta", 1)
        coefficient = beta_coefficient(alpha, beta)
    elif mode == SemiValueMode.Banzhaf:
        coefficient = banzhaf_coefficient
    else:
        raise ValueError(f"Unknown mode {mode}")
    coefficient = cast(SVCoefficient, coefficient)
    return semivalues(sampler_instance, u, coefficient, done, n_jobs=n_jobs, **kwargs)



import math
from time import sleep

import numpy as np
import pytest

from pydvl.utils import Status
from pydvl.value import ValuationResult
from pydvl.value.stopping import (
    AbsoluteStandardError,
    HistoryDeviation,
    MaxChecks,
    MaxTime,
    MaxUpdates,
    MinUpdates,
    StoppingCriterion,
    make_criterion,
)


def test_stopping_criterion():
    with pytest.raises(TypeError):
        StoppingCriterion()

    StoppingCriterion.__abstractmethods__ = set()
    done = StoppingCriterion()
    assert done.name == "StoppingCriterion"
    assert done.modify_result is True


def test_stopping_criterion_composition():
    StoppingCriterion.__abstractmethods__ = set()

    c = Status.Converged
    p = Status.Pending
    f = Status.Failed

    class C(StoppingCriterion):
        def _check(self, result: ValuationResult) -> Status:
            return c

    class P(StoppingCriterion):
        def _check(self, result: ValuationResult) -> Status:
            return p

    class F(StoppingCriterion):
        def _check(self, result: ValuationResult) -> Status:
            return f

    v = ValuationResult.empty()

    assert (~C())(v) == f
    assert (~P())(v) == c
    assert (~F())(v) == c

    assert (C() & C())(v) == c
    assert (P() | P())(v) == p

    assert (C() & P())(v) == (c & p)
    assert (C() | P())(v) == (c | p)

    assert (C() & C() & C())(v) == c
    assert (P() | P() | P())(v) == p

    assert (C() & P()).name == "Composite StoppingCriterion: C AND P"
    assert (C() | P()).name == "Composite StoppingCriterion: C OR P"
    assert (~C()).name == "Composite StoppingCriterion: NOT C"
    assert (~P()).name == "Composite StoppingCriterion: NOT P"


def test_make_criterion():
    def always_converged(result: ValuationResult) -> Status:
        return Status.Converged

    def always_pending(result: ValuationResult) -> Status:
        return Status.Pending

    def always_failed(result: ValuationResult) -> Status:
        return Status.Failed

    v = ValuationResult.empty()

    C = make_criterion(always_converged)
    P = make_criterion(always_pending)
    F = make_criterion(always_failed)

    assert C()(v) == Status.Converged
    assert P()(v) == Status.Pending
    assert F()(v) == Status.Failed

    assert C().name == "always_converged"
    assert P().name == "always_pending"
    assert F().name == "always_failed"

    assert (~C())(v) == Status.Failed
    assert (~P())(v) == Status.Converged
    assert (~F())(v) == Status.Converged

    assert (C() & C())(v) == Status.Converged
    assert (P() | P())(v) == Status.Pending
    assert (F() & F())(v) == Status.Failed
    assert (C() | F())(v) == Status.Converged


def test_minmax_updates():
    maxstop = MaxUpdates(10)
    assert maxstop.name == "MaxUpdates"
    v = ValuationResult.from_random(5)
    v._counts = np.zeros(5)
    assert maxstop(v) == Status.Pending
    v._counts += np.ones(5) * 9
    assert maxstop(v) == Status.Pending
    v._counts[0] += 1
    assert maxstop(v) == Status.Converged

    minstop = MinUpdates(10)
    assert minstop.name == "MinUpdates"
    v._counts = np.zeros(5)
    assert minstop(v) == Status.Pending
    v._counts += np.ones(5) * 9
    assert minstop(v) == Status.Pending
    v._counts[0] += 1
    assert minstop(v) == Status.Pending
    v._counts += np.ones(5)
    assert minstop(v) == Status.Converged


def test_max_time():
    v = ValuationResult.from_random(5)
    done = MaxTime(0.3)
    assert done(v) == Status.Pending
    sleep(0.3)
    assert done(v) == Status.Converged


@pytest.mark.parametrize("n_steps", [1, 42, 100])
@pytest.mark.parametrize("rtol", [0.01, 0.05])
def test_history_deviation(n_steps, rtol):
    """Values are equal and set to 1/t. The criterion will be fulfilled after
    t > (1+1/rtol) * n_steps iterations.
    """
    n = 5
    done = HistoryDeviation(n_steps=n_steps, rtol=rtol)
    threshold = math.ceil((1 + 1 / rtol) * n_steps)
    for t in range(1, threshold):
        v = ValuationResult(values=np.ones(n) / t, counts=np.full(n, t))
        assert done(v) == Status.Pending

    # FIXME: rounding errors mean that the threshold is not exactly as computed,
    #  but might be off by 1, so we check a couple of iterations to be sure that
    #  this works for any choice of n_steps and rtol
    status = Status.Pending
    for t in range(threshold, threshold + 2):
        v = ValuationResult(values=np.ones(n) / t, counts=np.full(n, t))
        status |= done(v)

    assert status == Status.Converged


def test_standard_error():
    """Test the AbsoluteStandardError stopping criterion."""
    eps = 0.1
    n = 5

    done = AbsoluteStandardError(threshold=eps, fraction=1.0, burn_in=0)

    # Trivial case: no variance.
    v = ValuationResult(values=np.ones(n), variances=np.zeros(n))
    assert done(v)

    # Reduce the variance until the criterion is triggered.
    v = ValuationResult(values=np.ones(n), variances=np.ones(n))
    assert not done(v)

    # One value is being left out
    for _ in range(10):
        for idx in range(1, n):
            v.update(idx, 1)
    assert not done(v)

    # Update the final value
    for _ in range(10):
        v.update(0, 1)
    assert done(v)


def test_max_checks():
    """Test the MaxChecks stopping criterion."""
    v = ValuationResult.from_random(size=5)

    done = MaxChecks(None)
    for _ in range(10):
        assert not done(v)

    done = MaxChecks(5)
    for _ in range(5):
        assert not done(v)
    assert done(v)



"""
This package contains algorithms for the computation of the influence function.

.. warning::
   Much of the code in this package is experimental or untested and is subject
   to modification. In particular, the package structure and basic API will
   probably change.

"""
from .general import *



import pytest
from numpy.testing import assert_almost_equal

from pydvl.value.least_core import exact_least_core
from tests.value import check_total_value, check_values


@pytest.mark.parametrize(
    "test_utility",
    [
        ("miner", {"n_miners": 3}),
        ("miner", {"n_miners": 4}),
        ("gloves", {"left": 1, "right": 1}),
        ("gloves", {"left": 2, "right": 1}),
        ("gloves", {"left": 1, "right": 2}),
    ],
    indirect=True,
)
@pytest.mark.parametrize("non_negative_subsidy", (True, False))
def test_naive_least_core(test_utility, non_negative_subsidy):
    u, exact_values = test_utility
    values = exact_least_core(
        u, non_negative_subsidy=non_negative_subsidy, progress=False
    )
    check_total_value(u, values)
    if non_negative_subsidy:
        check_values(values, exact_values)
        # Sometimes the subsidy is negative but really close to zero
        # Due to numerical errors
        if values.subsidy < 0:
            assert_almost_equal(values.subsidy, 0.0, decimal=5)
    else:
        check_values(values, exact_values, extra_values_names=["subsidy"])



import pytest

from pydvl.value.loo import naive_loo

from .. import check_total_value, check_values


@pytest.mark.parametrize("num_samples", [10, 100])
def test_naive_loo(num_samples, analytic_loo):
    """Compares the naive loo with analytic values in a dummy model"""
    u, exact_values = analytic_loo
    values = naive_loo(u, progress=False)
    check_total_value(u, values, rtol=0.1)
    check_values(values, exact_values, rtol=0.1)



import pytest

from pydvl.utils import Status
from pydvl.value.least_core.common import lc_solve_problem, lc_solve_problems
from pydvl.value.least_core.naive import lc_prepare_problem

from .. import check_values


@pytest.mark.parametrize(
    "test_utility",
    [("miner", {"n_miners": 5})],
    indirect=True,
)
def test_lc_solve_problems(test_utility, n_jobs, parallel_config):
    """Test solving LeastCoreProblems in parallel."""

    u, exact_values = test_utility
    n_problems = n_jobs
    problem = lc_prepare_problem(u)
    solutions = lc_solve_problems(
        [problem] * n_problems,
        u,
        algorithm="test_lc",
        n_jobs=n_jobs,
        config=parallel_config,
    )
    assert len(solutions) == n_problems

    for solution in solutions:
        assert solution.status == Status.Converged
        check_values(solution, exact_values, rtol=0.01)

        check = lc_solve_problem(problem, u=u, algorithm="test_lc")
        assert check.status == Status.Converged
        check_values(solution, check, rtol=0.01)



r"""
Monte Carlo approximations to Shapley Data values.

.. warning::
   You probably want to use the common interface provided by
   :func:`~pydvl.value.shapley.compute_shapley_values` instead of directly using
   the functions in this module.

Because exact computation of Shapley values requires $\mathcal{O}(2^n)$
re-trainings of the model, several Monte Carlo approximations are available.
The first two sample from the powerset of the training data directly:
:func:`combinatorial_montecarlo_shapley` and :func:`owen_sampling_shapley`. The
latter uses a reformulation in terms of a continuous extension of the utility.

Alternatively, employing another reformulation of the expression above as a sum
over permutations, one has the implementation in
:func:`permutation_montecarlo_shapley`, or using an early stopping strategy to
reduce computation :func:`truncated_montecarlo_shapley`.

.. seealso::
   It is also possible to use :func:`~pydvl.value.shapley.gt.group_testing_shapley`
   to reduce the number of evaluations of the utility. The method is however
   typically outperformed by others in this module.

.. seealso::
   Additionally, you can consider grouping your data points using
   :class:`~pydvl.utils.dataset.GroupedDataset` and computing the values of the
   groups instead. This is not to be confused with "group testing" as
   implemented in :func:`~pydvl.value.shapley.gt.group_testing_shapley`: any of
   the algorithms mentioned above, including Group Testing, can work to valuate
   groups of samples as units.
"""
import logging
import math
import operator
from functools import reduce
from itertools import cycle, takewhile
from typing import Sequence

import numpy as np
from numpy.typing import NDArray
from tqdm import tqdm

from pydvl.utils.config import ParallelConfig
from pydvl.utils.numeric import random_powerset
from pydvl.utils.parallel import MapReduceJob
from pydvl.utils.utility import Utility
from pydvl.value.result import ValuationResult
from pydvl.value.shapley.truncated import NoTruncation, TruncationPolicy
from pydvl.value.stopping import StoppingCriterion

logger = logging.getLogger(__name__)

__all__ = ["permutation_montecarlo_shapley", "combinatorial_montecarlo_shapley"]


def _permutation_montecarlo_shapley(
    u: Utility,
    *,
    done: StoppingCriterion,
    truncation: TruncationPolicy,
    algorithm_name: str = "permutation_montecarlo_shapley",
    progress: bool = False,
    job_id: int = 1,
) -> ValuationResult:
    """Helper function for :func:`permutation_montecarlo_shapley`.

    Computes marginal utilities of each training sample in
    :obj:`pydvl.utils.utility.Utility.data` by iterating through randomly
    sampled permutations.

    :param u: Utility object with model, data, and scoring function
    :param done: Check on the results which decides when to stop
    :param truncation: A callable which decides whether to interrupt
        processing a permutation and set all subsequent marginals to zero.
    :param algorithm_name: For the results object. Used internally by different
        variants of Shapley using this subroutine
    :param progress: Whether to display progress bars for each job.
    :param job_id: id to use for reporting progress (e.g. to place progres bars)
    :return: An object with the results
    """
    result = ValuationResult.zeros(
        algorithm=algorithm_name, indices=u.data.indices, data_names=u.data.data_names
    )

    pbar = tqdm(disable=not progress, position=job_id, total=100, unit="%")
    while not done(result):
        pbar.n = 100 * done.completion()
        pbar.refresh()
        prev_score = 0.0
        permutation = np.random.permutation(u.data.indices)
        permutation_done = False
        truncation.reset()
        for i, idx in enumerate(permutation):
            if permutation_done:
                score = prev_score
            else:
                score = u(permutation[: i + 1])
            marginal = score - prev_score
            result.update(idx, marginal)
            prev_score = score
            if not permutation_done and truncation(i, score):
                permutation_done = True
    return result


def permutation_montecarlo_shapley(
    u: Utility,
    done: StoppingCriterion,
    *,
    truncation: TruncationPolicy = NoTruncation(),
    n_jobs: int = 1,
    config: ParallelConfig = ParallelConfig(),
    progress: bool = False,
) -> ValuationResult:
    r"""Computes an approximate Shapley value by sampling independent index
    permutations to approximate the sum:

    $$
    v_u(x_i) = \frac{1}{n!} \sum_{\sigma \in \Pi(n)}
    \tilde{w}( | \sigma_{:i} | )[u(\sigma_{:i} \cup \{i\}) − u(\sigma_{:i})],
    $$

    where $\sigma_{:i}$ denotes the set of indices in permutation sigma before the
    position where $i$ appears (see :ref:`data valuation` for details).

    :param u: Utility object with model, data, and scoring function.
    :param done: function checking whether computation must stop.
    :param truncation: An optional callable which decides whether to
        interrupt processing a permutation and set all subsequent marginals to
        zero. Typically used to stop computation when the marginal is small.
    :param n_jobs: number of jobs across which to distribute the computation.
    :param config: Object configuring parallel computation, with cluster
        address, number of cpus, etc.
    :param progress: Whether to display progress bars for each job.
    :return: Object with the data values.
    """

    map_reduce_job: MapReduceJob[Utility, ValuationResult] = MapReduceJob(
        u,
        map_func=_permutation_montecarlo_shapley,
        reduce_func=lambda results: reduce(operator.add, results),
        map_kwargs=dict(
            algorithm_name="permutation_montecarlo_shapley",
            done=done,
            truncation=truncation,
            progress=progress,
        ),
        config=config,
        n_jobs=n_jobs,
    )
    return map_reduce_job()


def _combinatorial_montecarlo_shapley(
    indices: Sequence[int],
    u: Utility,
    done: StoppingCriterion,
    *,
    progress: bool = False,
    job_id: int = 1,
) -> ValuationResult:
    """Helper function for :func:`combinatorial_montecarlo_shapley`.

    This is the code that is sent to workers to compute values using the
    combinatorial definition.

    :param indices: Indices of the samples to compute values for.
    :param u: Utility object with model, data, and scoring function
    :param done: Check on the results which decides when to stop sampling
        subsets for an index.
    :param progress: Whether to display progress bars for each job.
    :param job_id: id to use for reporting progress
    :return: A tuple of ndarrays with estimated values and standard errors
    """
    n = len(u.data)

    # Correction coming from Monte Carlo integration so that the mean of the
    # marginals converges to the value: the uniform distribution over the
    # powerset of a set with n-1 elements has mass 2^{n-1} over each subset. The
    # additional factor n corresponds to the one in the Shapley definition
    correction = 2 ** (n - 1) / n
    result = ValuationResult.zeros(
        algorithm="combinatorial_montecarlo_shapley",
        indices=np.array(indices, dtype=np.int_),
        data_names=[u.data.data_names[i] for i in indices],
    )

    repeat_indices = takewhile(lambda _: not done(result), cycle(indices))
    pbar = tqdm(disable=not progress, position=job_id, total=100, unit="%")
    for idx in repeat_indices:
        pbar.n = 100 * done.completion()
        pbar.refresh()
        # Randomly sample subsets of full dataset without idx
        subset = np.setxor1d(u.data.indices, [idx], assume_unique=True)
        s = next(random_powerset(subset, n_samples=1))
        marginal = (u({idx}.union(s)) - u(s)) / math.comb(n - 1, len(s))
        result.update(idx, correction * marginal)

    return result


def combinatorial_montecarlo_shapley(
    u: Utility,
    done: StoppingCriterion,
    *,
    n_jobs: int = 1,
    config: ParallelConfig = ParallelConfig(),
    progress: bool = False,
) -> ValuationResult:
    r"""Computes an approximate Shapley value using the combinatorial
    definition:

    $$v_u(i) = \frac{1}{n} \sum_{S \subseteq N \setminus \{i\}}
    \binom{n-1}{ | S | }^{-1} [u(S \cup \{i\}) − u(S)]$$

    This consists of randomly sampling subsets of the power set of the training
    indices in :attr:`~pydvl.utils.utility.Utility.data`, and computing their
    marginal utilities. See :ref:`data valuation` for details.

    Note that because sampling is done with replacement, the approximation is
    poor even for $2^{m}$ subsets with $m>n$, even though there are $2^{n-1}$
    subsets for each $i$. Prefer
    :func:`~pydvl.shapley.montecarlo.permutation_montecarlo_shapley`.

    Parallelization is done by splitting the set of indices across processes and
    computing the sum over subsets $S \subseteq N \setminus \{i\}$ separately.

    :param u: Utility object with model, data, and scoring function
    :param done: Stopping criterion for the computation.
    :param n_jobs: number of parallel jobs across which to distribute the
        computation. Each worker receives a chunk of
        :attr:`~pydvl.utils.dataset.Dataset.indices`
    :param config: Object configuring parallel computation, with cluster
        address, number of cpus, etc.
    :param progress: Whether to display progress bars for each job.
    :return: Object with the data values.
    """

    map_reduce_job: MapReduceJob[NDArray, ValuationResult] = MapReduceJob(
        u.data.indices,
        map_func=_combinatorial_montecarlo_shapley,
        reduce_func=lambda results: reduce(operator.add, results),
        map_kwargs=dict(u=u, done=done, progress=progress),
        n_jobs=n_jobs,
        config=config,
    )
    return map_reduce_job()



import inspect
from functools import singledispatch, update_wrapper
from itertools import accumulate, repeat
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

import numpy as np
import ray
from numpy.typing import NDArray
from ray import ObjectRef

from ..config import ParallelConfig
from ..types import maybe_add_argument
from .backend import init_parallel_backend

__all__ = ["MapReduceJob"]

T = TypeVar("T")
R = TypeVar("R")
Identity = lambda x, *args, **kwargs: x

MapFunction = Callable[..., R]
ReduceFunction = Callable[[List[R]], R]
ChunkifyInputType = Union[NDArray[T], Sequence[T], T]


def _wrap_func_with_remote_args(func: Callable, *, timeout: Optional[float] = None):
    def wrapper(*args, **kwargs):
        args = list(args)
        for i, v in enumerate(args[:]):
            args[i] = _get_value(v, timeout=timeout)
        for k, v in kwargs.items():
            kwargs[k] = _get_value(v, timeout=timeout)
        return func(*args, **kwargs)

    try:
        inspect.signature(func)
        wrapper = update_wrapper(wrapper, func)
    except ValueError:
        # Doing it manually here because using update_wrapper from functools
        # on numpy functions doesn't work with ray for some unknown reason.
        wrapper.__name__ = func.__name__
        wrapper.__qualname__ = func.__qualname__
        wrapper.__doc__ = func.__doc__
    return wrapper


@singledispatch
def _get_value(v: Any, *, timeout: Optional[float] = None) -> Any:
    return v


@_get_value.register
def _(v: ObjectRef, *, timeout: Optional[float] = None) -> Any:
    return ray.get(v, timeout=timeout)


@_get_value.register
def _(v: np.ndarray, *, timeout: Optional[float] = None) -> NDArray:
    return v


# Careful to use list as hint. The dispatch does not work with typing generics
@_get_value.register
def _(v: list, *, timeout: Optional[float] = None) -> List[Any]:
    return [_get_value(x, timeout=timeout) for x in v]


class MapReduceJob(Generic[T, R]):
    """Takes an embarrassingly parallel fun and runs it in ``n_jobs`` parallel
    jobs, splitting the data evenly into a number of chunks equal to the number of jobs.

    Typing information for objects of this class requires the type of the inputs
    that are split for ``map_func`` and the type of its output.

    :param inputs: The input that will be split and passed to `map_func`.
        if it's not a sequence object. It will be repeat ``n_jobs`` number of times.
    :param map_func: Function that will be applied to the input chunks in each job.
    :param reduce_func: Function that will be applied to the results of
        ``map_func`` to reduce them.
    :param map_kwargs: Keyword arguments that will be passed to ``map_func`` in
        each job. Alternatively, one can use ``itertools.partial``.
    :param reduce_kwargs: Keyword arguments that will be passed to ``reduce_func``
        in each job. Alternatively, one can use :func:`itertools.partial`.
    :param config: Instance of :class:`~pydvl.utils.config.ParallelConfig`
        with cluster address, number of cpus, etc.
    :param n_jobs: Number of parallel jobs to run. Does not accept 0
    :param timeout: Amount of time in seconds to wait for remote results before
        ... TODO
    :param max_parallel_tasks: Maximum number of jobs to start in parallel. Any
        tasks above this number won't be submitted to the backend before some
        are done. This is to avoid swamping the work queue. Note that tasks have
        a low memory footprint, so this is probably not a big concern, except
        in the case of an infinite stream (not the case for MapReduceJob). See
        https://docs.ray.io/en/latest/ray-core/patterns/limit-pending-tasks.html

    :Examples:

    A simple usage example with 2 jobs:

    >>> from pydvl.utils.parallel import MapReduceJob
    >>> import numpy as np
    >>> map_reduce_job: MapReduceJob[np.ndarray, np.ndarray] = MapReduceJob(
    ...     np.arange(5),
    ...     map_func=np.sum,
    ...     reduce_func=np.sum,
    ...     n_jobs=2,
    ... )
    >>> map_reduce_job()
    10

    When passed a single object as input, it will be repeated for each job:

    >>> from pydvl.utils.parallel import MapReduceJob
    >>> import numpy as np
    >>> map_reduce_job: MapReduceJob[int, np.ndarray] = MapReduceJob(
    ...     5,
    ...     map_func=lambda x: np.array([x]),
    ...     reduce_func=np.sum,
    ...     n_jobs=4,
    ... )
    >>> map_reduce_job()
    20
    """

    def __init__(
        self,
        inputs: Union[Sequence[T], T],
        map_func: MapFunction[R],
        reduce_func: Optional[ReduceFunction[R]] = None,
        map_kwargs: Optional[Dict] = None,
        reduce_kwargs: Optional[Dict] = None,
        config: ParallelConfig = ParallelConfig(),
        *,
        n_jobs: int = -1,
        timeout: Optional[float] = None,
        max_parallel_tasks: Optional[int] = None,
    ):
        self.config = config
        parallel_backend = init_parallel_backend(self.config)
        self.parallel_backend = parallel_backend

        self.timeout = timeout

        self._n_jobs = 1
        # This uses the setter defined below
        self.n_jobs = n_jobs

        self.max_parallel_tasks = max_parallel_tasks

        self.inputs_ = inputs

        if reduce_func is None:
            reduce_func = Identity

        if map_kwargs is None:
            self.map_kwargs = dict()
        else:
            self.map_kwargs = {
                k: self.parallel_backend.put(v) for k, v in map_kwargs.items()
            }

        if reduce_kwargs is None:
            self.reduce_kwargs = dict()
        else:
            self.reduce_kwargs = {
                k: self.parallel_backend.put(v) for k, v in reduce_kwargs.items()
            }

        self._map_func = maybe_add_argument(map_func, "job_id")
        self._reduce_func = reduce_func

    def __call__(
        self,
    ) -> R:
        map_results = self.map(self.inputs_)
        reduce_results = self.reduce(map_results)
        return reduce_results

    def map(self, inputs: Union[Sequence[T], T]) -> List["ObjectRef[R]"]:
        """Splits the input data into chunks and calls a wrapped :func:`map_func` on them."""
        map_results: List["ObjectRef[R]"] = []

        map_func = self._wrap_function(self._map_func)

        total_n_jobs = 0
        total_n_finished = 0

        chunks = self._chunkify(inputs, n_chunks=self.n_jobs)

        for j, next_chunk in enumerate(chunks):
            result = map_func(next_chunk, job_id=j, **self.map_kwargs)
            map_results.append(result)
            total_n_jobs += 1

            total_n_finished = self._backpressure(
                map_results,
                n_dispatched=total_n_jobs,
                n_finished=total_n_finished,
            )
        return map_results

    def reduce(self, chunks: List["ObjectRef[R]"]) -> R:
        """Reduces the resulting chunks from a call to :meth:`~pydvl.utils.parallel.map_reduce.MapReduceJob.map`
        by passing them to a wrapped :func:`reduce_func`."""
        reduce_func = self._wrap_function(self._reduce_func)

        reduce_result = reduce_func(chunks, **self.reduce_kwargs)
        result = self.parallel_backend.get(reduce_result, timeout=self.timeout)
        return result  # type: ignore

    def _wrap_function(self, func: Callable, **kwargs) -> Callable:
        """Wraps a function with a timeout and remote arguments and puts it on
        the remote backend.

        :param func: Function to wrap
        :param kwargs: Additional keyword arguments to pass to the backend
            wrapper. These are *not* arguments for the wrapped function.
        :return: Remote function that can be called with the same arguments as
            the wrapped function. Depending on the backend, this may simply be
            the function itself.
        """
        return self.parallel_backend.wrap(
            _wrap_func_with_remote_args(func, timeout=self.timeout), **kwargs
        )

    def _backpressure(
        self, jobs: List[ObjectRef], n_dispatched: int, n_finished: int
    ) -> int:
        """This is used to limit the number of concurrent tasks.
        If :attr:`~pydvl.utils.parallel.map_reduce.MapReduceJob.max_parallel_tasks` is None then this function
        is a no-op that simply returns 0.

        See https://docs.ray.io/en/latest/ray-core/patterns/limit-pending-tasks.html
        :param jobs:
        :param n_dispatched:
        :param n_finished:
        :return:
        """
        if self.max_parallel_tasks is None:
            return 0
        while (n_in_flight := n_dispatched - n_finished) > self.max_parallel_tasks:
            wait_for_num_jobs = n_in_flight - self.max_parallel_tasks
            finished_jobs, _ = self.parallel_backend.wait(
                jobs,
                num_returns=wait_for_num_jobs,
                timeout=10,  # FIXME make parameter?
            )
            n_finished += len(finished_jobs)
        return n_finished

    def _chunkify(self, data: ChunkifyInputType, n_chunks: int) -> List["ObjectRef[T]"]:
        """If data is a Sequence, it splits it into Sequences of size `n_chunks` for each job that we call chunks.
        If instead data is an `ObjectRef` instance, then it yields it repeatedly `n_chunks` number of times.
        """
        if n_chunks <= 0:
            raise ValueError("Number of chunks should be greater than 0")

        if n_chunks == 1:
            data_id = self.parallel_backend.put(data)
            return [data_id]

        try:
            # This is used as a check to determine whether data is iterable or not
            # if it's the former, then the value will be used to determine the chunk indices.
            n = len(data)
        except TypeError:
            data_id = self.parallel_backend.put(data)
            return list(repeat(data_id, times=n_chunks))
        else:
            # This is very much inspired by numpy's array_split function
            # The difference is that it only uses built-in functions
            # and does not convert the input data to an array
            chunk_size, remainder = divmod(n, n_chunks)
            chunk_indices = tuple(
                accumulate(
                    [0]
                    + remainder * [chunk_size + 1]
                    + (n_chunks - remainder) * [chunk_size]
                )
            )

            chunks = []

            for start_index, end_index in zip(chunk_indices[:-1], chunk_indices[1:]):
                if start_index >= end_index:
                    break
                chunk_id = self.parallel_backend.put(data[start_index:end_index])
                chunks.append(chunk_id)

            return chunks

    @property
    def n_jobs(self) -> int:
        """Effective number of jobs according to the used ParallelBackend instance."""
        return self._n_jobs

    @n_jobs.setter
    def n_jobs(self, value: int):
        self._n_jobs = self.parallel_backend.effective_n_jobs(value)



"""
Contains tests for LinearRegression, BinaryLogisticRegression as well as TwiceDifferentiable modules and
its associated gradient and matrix vector product calculations. Note that there is no test for the neural network
module.
"""

import itertools
from typing import List, Tuple

import numpy as np
import pytest

from pydvl.utils import (
    linear_regression_analytical_derivative_d2_theta,
    linear_regression_analytical_derivative_d_theta,
    linear_regression_analytical_derivative_d_x_d_theta,
)

try:
    import torch.nn.functional as F

    from pydvl.influence.frameworks import TorchTwiceDifferentiable
    from pydvl.influence.model_wrappers import TorchLinearRegression
except ImportError:
    pass


class ModelTestSettings:
    DATA_OUTPUT_NOISE: float = 0.01
    ACCEPTABLE_ABS_TOL_MODEL: float = (
        0.04  # TODO: Reduce bound if tests are running with fixed seeds.
    )
    ACCEPTABLE_ABS_TOL_DERIVATIVE: float = 1e-5

    TEST_CONDITION_NUMBERS: List[int] = [5]
    TEST_SET_SIZE: List[int] = [20]
    TRAINING_SET_SIZE: List[int] = [500]
    PROBLEM_DIMENSIONS: List[Tuple[int, int]] = [
        (2, 2),
        (5, 10),
        (10, 5),
        (10, 10),
    ]


test_cases_linear_regression_fit = list(
    itertools.product(
        ModelTestSettings.TRAINING_SET_SIZE,
        ModelTestSettings.TEST_SET_SIZE,
        ModelTestSettings.PROBLEM_DIMENSIONS,
        ModelTestSettings.TEST_CONDITION_NUMBERS,
    )
)

test_cases_logistic_regression_fit = list(
    itertools.product(
        ModelTestSettings.TRAINING_SET_SIZE,
        ModelTestSettings.TEST_SET_SIZE,
        [(1, 3), (1, 7), (1, 20)],
        ModelTestSettings.TEST_CONDITION_NUMBERS,
    )
)

test_cases_linear_regression_derivatives = list(
    itertools.product(
        ModelTestSettings.TRAINING_SET_SIZE,
        ModelTestSettings.PROBLEM_DIMENSIONS,
        ModelTestSettings.TEST_CONDITION_NUMBERS,
    )
)


def lmb_fit_test_case_to_str(packed_i_test_case):
    i, test_case = packed_i_test_case
    return f"Problem #{i} of dimension {test_case[2]} with train size {test_case[0]}, test size {test_case[1]} and condition number {test_case[3]}"


def lmb_correctness_test_case_to_str(packed_i_test_case):
    i, test_case = packed_i_test_case
    return f"Problem #{i} of dimension {test_case[1]} with train size {test_case[0]} and condition number {test_case[2]}"


fit_test_case_ids = list(
    map(
        lmb_fit_test_case_to_str,
        zip(
            range(len(test_cases_linear_regression_fit)),
            test_cases_linear_regression_fit,
        ),
    )
)
correctness_test_case_ids = list(
    map(
        lmb_correctness_test_case_to_str,
        zip(
            range(len(test_cases_linear_regression_derivatives)),
            test_cases_linear_regression_derivatives,
        ),
    )
)


@pytest.mark.torch
@pytest.mark.parametrize(
    "train_set_size,problem_dimension,condition_number",
    test_cases_linear_regression_derivatives,
    ids=correctness_test_case_ids,
)
def test_linear_regression_model_grad(
    train_set_size: int,
    condition_number: float,
    linear_model: Tuple[np.ndarray, np.ndarray],
):
    # some settings
    A, b = linear_model
    output_dimension, input_dimension = tuple(A.shape)

    # generate datasets
    data_model = lambda x: np.random.normal(
        x @ A.T + b, ModelTestSettings.DATA_OUTPUT_NOISE
    )
    train_x = np.random.uniform(size=[train_set_size, input_dimension])
    train_y = data_model(train_x)

    model = TorchLinearRegression(input_dimension, output_dimension, init=linear_model)
    loss = F.mse_loss
    mvp_model = TorchTwiceDifferentiable(model=model, loss=loss)

    train_grads_analytical = 2 * linear_regression_analytical_derivative_d_theta(
        (A, b), train_x, train_y
    )
    train_grads_autograd = mvp_model.split_grad(train_x, train_y)
    train_grads_max_diff = np.max(np.abs(train_grads_analytical - train_grads_autograd))
    assert (
        train_grads_max_diff < ModelTestSettings.ACCEPTABLE_ABS_TOL_DERIVATIVE
    ), "training set produces wrong gradients."


@pytest.mark.torch
@pytest.mark.parametrize(
    "train_set_size,problem_dimension,condition_number",
    test_cases_linear_regression_derivatives,
    ids=correctness_test_case_ids,
)
def test_linear_regression_model_hessian(
    train_set_size: int,
    condition_number: float,
    linear_model: Tuple[np.ndarray, np.ndarray],
):
    # some settings
    A, b = linear_model
    output_dimension, input_dimension = tuple(A.shape)

    # generate datasets
    data_model = lambda x: np.random.normal(
        x @ A.T + b, ModelTestSettings.DATA_OUTPUT_NOISE
    )
    train_x = np.random.uniform(size=[train_set_size, input_dimension])
    train_y = data_model(train_x)
    model = TorchLinearRegression(input_dimension, output_dimension, init=linear_model)
    loss = F.mse_loss
    mvp_model = TorchTwiceDifferentiable(model=model, loss=loss)

    test_hessian_analytical = 2 * linear_regression_analytical_derivative_d2_theta(
        (A, b), train_x, train_y
    )
    grad_xy, _ = mvp_model.grad(train_x, train_y)
    estimated_hessian = mvp_model.mvp(
        grad_xy, np.eye((input_dimension + 1) * output_dimension)
    )
    test_hessian_max_diff = np.max(np.abs(test_hessian_analytical - estimated_hessian))
    assert (
        test_hessian_max_diff < ModelTestSettings.ACCEPTABLE_ABS_TOL_DERIVATIVE
    ), "Hessian was wrong."


@pytest.mark.torch
@pytest.mark.parametrize(
    "train_set_size,problem_dimension,condition_number",
    test_cases_linear_regression_derivatives,
    ids=correctness_test_case_ids,
)
def test_linear_regression_model_d_x_d_theta(
    train_set_size: int,
    condition_number: float,
    linear_model: Tuple[np.ndarray, np.ndarray],
):
    # some settings
    A, b = linear_model
    output_dimension, input_dimension = tuple(A.shape)

    # generate datasets
    data_model = lambda x: np.random.normal(
        x @ A.T + b, ModelTestSettings.DATA_OUTPUT_NOISE
    )
    train_x = np.random.uniform(size=[train_set_size, input_dimension])
    train_y = data_model(train_x)
    model = TorchLinearRegression(input_dimension, output_dimension, init=(A, b))
    loss = F.mse_loss
    mvp_model = TorchTwiceDifferentiable(model=model, loss=loss)

    test_derivative = 2 * linear_regression_analytical_derivative_d_x_d_theta(
        (A, b),
        train_x,
        train_y,
    )
    model_mvp = []
    for i in range(len(train_x)):
        grad_xy, tensor_x = mvp_model.grad(train_x[i], train_y[i])
        model_mvp.append(
            mvp_model.mvp(
                grad_xy,
                np.eye((input_dimension + 1) * output_dimension),
                backprop_on=tensor_x,
            )
        )
    estimated_derivative = np.stack(model_mvp, axis=0)
    test_hessian_max_diff = np.max(np.abs(test_derivative - estimated_derivative))
    assert (
        test_hessian_max_diff < ModelTestSettings.ACCEPTABLE_ABS_TOL_DERIVATIVE
    ), "Hessian was wrong."



import numpy as np
import pytest
import ray
from numpy.typing import NDArray
from ray.cluster_utils import Cluster
from sklearn.linear_model import LinearRegression
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import PolynomialFeatures

from pydvl.utils import Dataset, SupervisedModel, Utility
from pydvl.utils.config import ParallelConfig
from pydvl.utils.status import Status
from pydvl.value import ValuationResult

from . import polynomial


@pytest.fixture(scope="function")
def polynomial_dataset(coefficients: np.ndarray):
    """Coefficients must be for monomials of increasing degree"""
    from sklearn.utils import Bunch

    x = np.arange(-1, 1, 0.05)
    locs = polynomial(coefficients, x)
    y = np.random.normal(loc=locs, scale=0.3)
    db = Bunch()
    db.data, db.target = x.reshape(-1, 1), y
    poly = [f"{c} x^{i}" for i, c in enumerate(coefficients)]
    poly = " + ".join(poly)
    db.DESCR = f"$y \\sim N({poly}, 1)$"
    db.feature_names = ["x"]
    db.target_names = ["y"]
    return Dataset.from_sklearn(data=db, train_size=0.15), coefficients


@pytest.fixture(scope="function")
def polynomial_pipeline(coefficients):
    return make_pipeline(PolynomialFeatures(len(coefficients) - 1), LinearRegression())


@pytest.fixture(scope="function")
def dummy_utility(num_samples):
    # Indices match values
    x = np.arange(0, num_samples, 1).reshape(-1, 1)
    nil = np.zeros_like(x)
    data = Dataset(
        x, nil, nil, nil, feature_names=["x"], target_names=["y"], description="dummy"
    )

    class DummyModel(SupervisedModel):
        """Under this model each data point receives a score of index / max,
        assuming that the values of training samples match their indices.

        The utility of a set is the sum of the utilities.
        """

        def __init__(self, data: Dataset):
            self.m = max(data.x_train)
            self.utility = 0

        def fit(self, x: NDArray, y: NDArray):
            self.utility = np.sum(x) / self.m

        def predict(self, x: NDArray) -> NDArray:
            return x

        def score(self, x: NDArray, y: NDArray) -> float:
            return self.utility

    return Utility(
        DummyModel(data),
        data,
        score_range=(0, x.sum() / x.max()),
        catch_errors=False,
        show_warnings=True,
        enable_cache=False,
    )


@pytest.fixture(scope="function")
def analytic_shapley(dummy_utility):
    """Scores are i/n, so v(i) = 1/n! Σ_π [U(S^π + {i}) - U(S^π)] = i/n"""

    m = float(max(dummy_utility.data.x_train))
    values = np.array([i / m for i in dummy_utility.data.indices])
    result = ValuationResult(
        algorithm="exact",
        values=values,
        variances=np.zeros_like(values),
        data_names=dummy_utility.data.indices,
        status=Status.Converged,
    )
    return dummy_utility, result


@pytest.fixture(scope="function")
def analytic_banzhaf(dummy_utility):
    """Scores are i/n, so
    v(i) = 1/2^{n-1} Σ_{S_{-i}} [U(S + {i}) - U(S)] = i/n
    """

    m = float(max(dummy_utility.data.x_train))
    values = np.array([i / m for i in dummy_utility.data.indices])
    result = ValuationResult(
        algorithm="exact",
        values=values,
        variances=np.zeros_like(values),
        data_names=dummy_utility.data.indices,
        status=Status.Converged,
    )
    return dummy_utility, result


@pytest.fixture(scope="function")
def linear_shapley(linear_dataset, scorer, n_jobs):
    u = Utility(
        LinearRegression(), data=linear_dataset, scorer=scorer, enable_cache=False
    )

    from pydvl.value.shapley.naive import combinatorial_exact_shapley

    exact_values = combinatorial_exact_shapley(u, progress=False, n_jobs=n_jobs)
    return u, exact_values


@pytest.fixture(scope="module", params=["sequential", "ray-local", "ray-external"])
def parallel_config(request):
    if request.param == "sequential":
        yield ParallelConfig(backend=request.param)
    elif request.param == "ray-local":
        yield ParallelConfig(backend="ray")
        ray.shutdown()
    elif request.param == "ray-external":
        # Starts a head-node for the cluster.
        cluster = Cluster(initialize_head=True, head_node_args={"num_cpus": 4})
        yield ParallelConfig(backend="ray", address=cluster.address)
        ray.shutdown()
        cluster.shutdown()
